import struct
from xml.dom import minidom

extensions = {
    # Unhashed or unknown extension
    0x00B1B104: "prop",
    0x00E6BCE5: "gmdl",
    0x011989B7: "plt", # Palette
    0x01AD2416: "creature_traits",
    0x01AD2417: "building_traits",
    0x01AD2418: "vehicle_traits",
    0x01C135DA: "gmsh",
    0x01C3C4B3: "trait_pill",
    0x0248F226: "css",
    0x024A0E52: "trigger",
    0x02523258: "formation",
    0x027C5CEF: "ttf", # NEW
    0x02D5C9AF: "summary",
    0x02D5C9B0: "summary_pill",
    0x02FAC0B6: "locale",
    0x030BDEE3: "pollen_metadata",
    0x0376C3DA: "hm",
    0x0472329B: "htra",
    0x065266B7: "xhtml", # NEW
    0x2F4E681C: "raster",
    0x2F7D0002: "jpeg",
    0x2F7D0004: "png",
    0x2F7D0005: "bmp",
    0x2F7D0006: "tga",
    0x2F7D0007: "gif",
    0x4AEB6BC6: "tlsa",
    0x7C19AA7A: "pctp",
    0xCF6C21B8: "xml", # NEW
    0xEFBDA3FF: "layout", # source format for .spui?

    # Hashed version of the extension
    0x12952634: "dat",
    0x1A99B06B: "bem",
    0x1E99B626: "bat",
    0x1F639D98: "xls",
    0x2399BE55: "bld", # building
    0x24682294: "vcl", # vehicle (land, sea, air)
    0x250FE9A2: "spui", # SPore User Interface
    0x25DF0112: "gait",
    0x2B6CAB5F: "txt",
    0x2B978C46: "crt", # creature
    0x376840D7: "vp6", # VP6 encoded video, see http://en.wikipedia.org/wiki/VP6
    0x37979F71: "cfg",
    0x3C77532E: "psd",
    0x3C7E0F63: "mcl", # muscle
    0x3D97A8E4: "cll", # cell
    0x3F9C28B5: "ani",
    0x438F6347: "flr", # flora
    0x448AE7E2: "hkx", # havok physics (or effect?)
    0x476A98C7: "ufo", # spaceship
    0x497767B9: "pfx", # particle effect
    0x5C74D18B: "density",
    0x617715C4: "py",
    0x617715D9: "pd", # sound related
    0x9B8E862F: "world",
    0xDFAD9F51: "cell",
    0xEE17C6AD: "animation",

    0x02393756: "cur",
    0x2F4E681B: "rw4",
    0x06EFC6AA: "package",
    0x01A527DB: "snr",
    0x03F51892: "children",
    0x022D2C83: "pdr",
    0x03055F61: "instrument",
    0x01EEF63A: "sns",
    0x02B9F662: "a.prop",
    0x0497925E: "mode",
    0x029E333B: "voice",
    0x042C9CBB: "sek",
    0x2E5A9763: "lru",
    0x02C9EFF2: "submix"
}

groupNames = {}
for item in minidom.parse("/Users/impiaaa/Downloads/Sources/Gibbed.Spore/bin/group_names.xml").getElementsByTagName('name'):
    groupNames[int(item.attributes["id"].value, 16)] = item.firstChild.data

def getExtensionFromId(id):
    if extensions.has_key(id): return extensions[id]
    else: return "0x%08X.cdata" % id

class DatabasePackedFileException(Exception):
    pass

class NotAPackageException(DatabasePackedFileException):
    pass

class UnsupportedPackageVersionException(DatabasePackedFileException):
    pass

class DatabaseIndex:
    def checkCompressed(self):
        self.compressedSize &= ~0x80000000
        if self.compressedFlags != 0 and self.compressedFlags != -1:
            raise ValueError("compressed flags (%04X)" % self.compressedFlags)
        self.compressed = self.compressedFlags == -1
    def __repr__(self):
        return "%s: 0x%08X, 0x%08X @ 0x%08X" % (self.__class__.__name__, self.typeId, self.groupId, self.instanceId)

class DatabasePackedFileHeader:
    def read(self, stream):
        fmt = 'll24xl4xl12xLl28x'
        self.majorVersion,\
        self.minorVersion,\
        self.indexCount,\
        self.indexSize,\
        self.always3,\
        self.indexOffset = struct.unpack(fmt, stream.read(struct.calcsize(fmt)))

class DatabaseBigPackageFileHeader:
    def read(self, stream):
        fmt = 'll24xlq4xLq56x'
        self.majorVersion,\
        self.minorVersion,\
        self.indexCount,\
        self.indexSize,\
        self.always3,\
        self.indexOffset = struct.unpack(fmt, stream.read(struct.calcsize(fmt)))

class Version:
    def __init__(self, majorVersion, minorVersion):
        self.major = majorVersion
        self.minor = minorVersion

class DatabasePackedFile:
    def read(self, stream):
        big = False
        magic = stream.read(4)
        if magic != "DBPF" and magic != "DBBF":
            raise NotAPackageException()
        if magic == "DBBF":
            big = True
            header = DatabaseBigPackageFileHeader()
            header.read(stream)
            if header.always3 != 3:
                raise Exception("the value in the DBBF header that is always 3 was not 3")
            # Nab useful stuff
            self.version = Version(header.majorVersion, header.minorVersion)
            indexCount = header.indexCount
            indexOffset = header.indexOffset
            indexSize = header.indexSize
        else:
            big = False
            header = DatabasePackedFileHeader()
            header.read(stream)
            if header.always3 != 3:
                raise Exception("the value in the DBPF header that is always 3 was not 3")
            # Nab useful stuff
            self.version = Version(header.majorVersion, header.minorVersion)
            indexCount = header.indexCount
            indexOffset = header.indexOffset
            indexSize = header.indexSize
        self.indices = []
        if indexCount > 0:
            # Read index
            stream.seek(indexOffset)
            fmt = 'l'
            indexHeaderValues = struct.unpack(fmt, stream.read(struct.calcsize(fmt)))[0]
            if indexHeaderValues < 4 or indexHeaderValues > 7:
                raise ValueError("don't know how to handle this index data")
            indexTypeId = 0xCAFEBABE
            # type id
            fmt = 'L'
            if (indexHeaderValues & (1 << 0)) == 1 << 0:
                indexTypeId = struct.unpack(fmt, stream.read(struct.calcsize(fmt)))[0]
            indexGroupId = 0xCAFEBABE
            # group id
            if (indexHeaderValues & (1 << 1)) == 1 << 1:
                indexGroupId = struct.unpack(fmt, stream.read(struct.calcsize(fmt)))[0]
            indexUnknown = 0xCAFEBABE
            # unknown value
            if (indexHeaderValues & (1 << 2)) == 1 << 2:
                indexGroupId = struct.unpack(fmt, stream.read(struct.calcsize(fmt)))[0]
            for i in xrange(indexCount):
                fmt = 'L'
                index = DatabaseIndex()
                if (indexHeaderValues & (1 << 0)) == 1 << 0:
                    index.typeId = indexTypeId
                else:
                    index.typeId = struct.unpack(fmt, stream.read(struct.calcsize(fmt)))[0]
                if (indexHeaderValues & (1 << 1)) == 1 << 1:
                    index.groupId = indexGroupId
                else:
                    index.groupId = struct.unpack(fmt, stream.read(struct.calcsize(fmt)))[0]
                if (indexHeaderValues & (1 << 2)) == 1 << 2:
                    index.unknown = indexUnknown
                else:
                    index.unknown = struct.unpack(fmt, stream.read(struct.calcsize(fmt)))[0]
                index.instanceId = struct.unpack(fmt, stream.read(struct.calcsize(fmt)))[0]
                if big:
                    fmt = 'Q'
                else:
                    fmt = 'L'
                index.offset = struct.unpack(fmt, stream.read(struct.calcsize(fmt)))[0]
                fmt = 'LLhH'
                index.compressedSize,\
                index.decompressedSize,\
                index.compressedFlags,\
                index.flags = struct.unpack(fmt, stream.read(struct.calcsize(fmt)))
                index.checkCompressed()
                self.indices += [index]

def readRefPackCompressionHeader(stream):
    header = [ord(c) for c in stream.read(2)]
    #stop = False
    # hdr & 0x3EFF) == 0x10FB
    if (header[0] & 0x3E) != 0x10 or (header[1] != 0xFB):
        # stream is not compressed
        #stop = True
        pass
    # read destination (uncompressed) length
    isLong = (header[0] & 0x80) != 0
    hasMore = (header[0] & 0x01) != 0
    data = [ord(c) for c in stream.read((4 if isLong else 3) * (2 if hasMore else 1))]
    realLength = (((data[0] << 8) + data[1]) << 8) + data[2]
    if isLong:
        realLength = (realLength << 8) + data[3]

def refPackDecompress(stream, compressedSize, decompressedSize):
    baseOffset = stream.tell()
    outputData = [0]*decompressedSize
    offset = 0
    readRefPackCompressionHeader(stream)
    while stream.tell() < baseOffset + compressedSize:
        stop = False
        plainSize = 0
        copySize = 0
        copyOffset = 0
        prefix = ord(stream.read(1))
        if prefix >= 0xC0:
            if prefix >= 0xE0:
                if prefix >= 0xFC:
                    plainSize = prefix & 3
                    stop = True
                else:
                    plainSize = ((prefix & 0x1F) + 1) * 4
            else:
                extra = [ord(c) for c in stream.read(3)]
                plainSize = prefix & 3
                copySize = (((prefix & 0x0C) << 6) | extra[2]) + 5
                copyOffset = (((((prefix & 0x10) << 4) | extra[0]) << 8) | extra[1]) + 1
        else:
            if prefix >= 0x80:
                extra = [ord(c) for c in stream.read(2)]
                plainSize = extra[0] >> 6
                copySize = (prefix & 0x3F) + 4
                copyOffset = (((extra[0] & 0x3F) << 8) | extra[1]) + 1
            else:
                extra = ord(stream.read(1))
                plainSize = prefix & 3
                copySize = ((prefix & 0x1C) >> 2) + 3
                copyOffset = (((prefix & 0x60) << 3) | extra) + 1
        if plainSize > 0:
            outputData[offset:offset+plainSize] = [ord(c) for c in stream.read(plainSize)]
            offset += plainSize
        if copySize > 0:
            for i in xrange(copySize):
                outputData[offset + i] = outputData[(offset - copyOffset) + i]
            offset += copySize
        if stop:
            break
    return outputData

def extractByInstance(dpf, f, instanceId):
    for index in dpf.indices:
        if index.instanceId != instanceId: continue
        if groupNames.has_key(index.groupId):
            gname = groupNames[index.groupId]
        else:
            gname = "0x%08X" % index.groupId
        out = open("/Users/impiaaa/Desktop/extracts/%s-0x%08X.%s" % (gname, index.instanceId, getExtensionFromId(index.typeId)), 'wb')
        f.seek(index.offset)
        try:
            if index.compressed:
                out.write(''.join([chr(c) for c in refPackDecompress(f, index.compressedSize, index.decompressedSize)]))
            else:
                out.write(f.read(index.compressedSize))
        except OverflowError:
            pass
        out.close()
    
if __name__ == '__main__':
    import os, sys
    if len(sys.argv) < 2:
        sys.stderr.write("Usage:\n")
        sys.stderr.write("python package.py <package file>")
        sys.exit(1)
    f = open(sys.argv[1], 'rb')
    packagename = os.path.split(sys.argv[1])[1]
    if not os.path.isdir(packagename):
        os.mkdir(packagename)
    dpf = DatabasePackedFile()
    dpf.read(f)
    for index in dpf.indices:
        if groupNames.has_key(index.groupId):
            dname = os.path.join(packagename, groupNames[index.groupId])
        else:
            dname = os.path.join(packagename, "0x%08X" % index.groupId)
        if not os.path.isdir(dname): os.mkdir(dname)
        out = open("%s/0x%08X.%s" % (dname, index.instanceId, getExtensionFromId(index.typeId)), 'wb')
        f.seek(index.offset)
        try:
            if index.compressed:
                out.write(''.join([chr(c) for c in refPackDecompress(f, index.compressedSize, index.decompressedSize)]))
            else:
                out.write(f.read(index.compressedSize))
        except OverflowError:
            pass
        out.close()
    f.close()
